@@ -1,6 +1,8 @@
 # gemato: OpenPGP verification support
-# (c) 2017-2023 Michał Górny
-# Licensed under the terms of 2-clause BSD license
+# (c) 2017-2024 Michał Górny
+# SPDX-License-Identifier: GPL-2.0-or-later
+
+from __future__ import annotations
 
 import base64
 import dataclasses
@@ -10,6 +12,7 @@ import enum
 import errno
 import hashlib
 import logging
+import math
 import os
 import os.path
 import shutil
@@ -48,6 +51,8 @@ except ImportError:
 GNUPG = os.environ.get('GNUPG', 'gpg')
 GNUPGCONF = os.environ.get('GNUPGCONF', 'gpgconf')
 
+LOGGER = logging.getLogger(__name__)
+
 
 class OpenPGPSignatureStatus(enum.Enum):
     GOOD = enum.auto()
@@ -68,6 +73,7 @@ class OpenPGPSignatureData:
     sig_status: typing.Optional[OpenPGPSignatureStatus] = None
     valid_sig: bool = False
     trusted_sig: bool = False
+    key_expiration: typing.Optional[datetime.datetime] = None
 
 
 class OpenPGPSignatureList(list[OpenPGPSignatureData]):
@@ -111,10 +117,9 @@ class SystemGPGEnvironment:
     (user's home directory or GNUPGHOME).
     """
 
-    __slots__ = ['debug']
-
-    def __init__(self, debug=False, proxy=None):
+    def __init__(self, debug=False, proxy=None, timeout=None):
         self.debug = debug
+        self._trusted_keys = set()
 
     def __enter__(self):
         return self
@@ -135,6 +140,60 @@ class SystemGPGEnvironment:
         raise NotImplementedError(
             'import_key() is not implemented by this OpenPGP provider')
 
+    def list_keys(self, key_ids: list[str] = []
+                  ) -> dict[str, list[tuple[str, str]]]:
+        """
+        List fingerprints and UIDs of specified keys or all keys in keyring
+
+        Returns a mapping from fingerprint (as a string) to an iterable
+        of UIDs.
+        """
+
+        exitst, out, err = self._spawn_gpg(
+            [GNUPG, '--batch', '--with-colons', '--list-keys', *key_ids],
+            raise_on_error=OpenPGPKeyListingError)
+
+        prev_pub = None
+        fpr = None
+        ret = {}
+        for line in out.splitlines():
+            # were we expecting a fingerprint?
+            if prev_pub is not None:
+                if line.startswith(b'fpr:'):
+                    fpr = line.split(b':')[9].decode('ASCII')
+                    if not fpr.endswith(prev_pub):
+                        raise OpenPGPKeyListingError(
+                            f'Incorrect fingerprint {fpr} for key '
+                            f'{prev_pub}')
+                    LOGGER.debug(
+                        f'list_keys(): fingerprint: {fpr}')
+                    ret[fpr] = []
+                    prev_pub = None
+                else:
+                    raise OpenPGPKeyListingError(
+                        f'No fingerprint in GPG output, instead got: '
+                        f'{line}')
+            elif line.startswith(b'pub:'):
+                # wait for the fingerprint
+                prev_pub = line.split(b':')[4].decode('ASCII')
+                LOGGER.debug(f'list_keys(): keyid: {prev_pub}')
+            elif line.startswith(b'uid:'):
+                if fpr is None:
+                    raise OpenPGPKeyListingError(
+                        f'UID without key in GPG output: {line}')
+                uid_split = line.split(b":", 10)
+                uid = uid_split[9]
+                # no creation date means missing/broken self-sig
+                if not uid_split[5]:
+                    LOGGER.debug(
+                        f"list_keys(): skipping UID with missing self-sig: "
+                        f"{fpr=}, {uid=!r}")
+                    continue
+                LOGGER.debug(f'list_keys(): UID: {uid}')
+                ret[fpr].append(uid)
+
+        return ret
+
     def refresh_keys(self, allow_wkd=True, keyserver=None):
         """
         Update the keys from their assigned keyservers. This should be called
@@ -210,7 +269,7 @@ class SystemGPGEnvironment:
                 sig_list[-1].fingerprint = spl[2].decode('utf8')
                 sig_list[-1].timestamp = (
                     self._parse_gpg_ts(spl[4].decode('utf8')))
-                sig_list[-1].expiration_timestamp = (
+                sig_list[-1].expire_timestamp = (
                     self._parse_gpg_ts(spl[5].decode('utf8')))
                 sig_list[-1].primary_key_fingerprint = spl[11].decode('utf8')
             elif line.startswith(b'[GNUPG:] TRUST_'):
@@ -220,17 +279,40 @@ class SystemGPGEnvironment:
                               b'TRUST_FULL',
                               b'TRUST_ULTIMATE'):
                     sig_list[-1].trusted_sig = True
+            elif line.startswith(b"[GNUPG:] KEYEXPIRED"):
+                # TODO: will the "correct" key be emitted last?
+                assert sig_list
+                spl = line.split(b" ", 3)
+                assert len(spl) >= 3
+                sig_list[-1].key_expiration = (
+                    self._parse_gpg_ts(spl[2].decode("utf8")))
 
         if not sig_list:
             raise OpenPGPUnknownSigFailure(
                 err.decode('utf8', errors='backslashreplace'))
 
-        # bad signature causes failure even without require_all_good
         for sig in sig_list:
+            # bad signature causes failure even without require_all_good
             if sig.sig_status == OpenPGPSignatureStatus.BAD:
                 raise OpenPGPVerificationFailure(
                     err.decode("utf8", errors="backslashreplace"), sig)
 
+            if sig.sig_status == OpenPGPSignatureStatus.EXPIRED_KEY:
+                # if the signature was done using a key that's expired *now*,
+                # check whether it was expired prior to the signature.
+                assert sig.timestamp is not None
+                assert sig.key_expiration is not None
+                if sig.timestamp < sig.key_expiration:
+                    sig.sig_status = OpenPGPSignatureStatus.GOOD
+                    # we need to check the trust explicitly since GPG
+                    # doesn't check trust on expired keys
+                    if sig.primary_key_fingerprint in self._trusted_keys:
+                        sig.trusted_sig = True
+            else:
+                # remove key_expiration if it the signature was not done
+                # by an expired key
+                sig.key_expiration = None
+
         if not require_all_good:
             if any(x.sig_status == OpenPGPSignatureStatus.GOOD and
                    x.valid_sig and x.trusted_sig for x in sig_list):
@@ -287,14 +369,15 @@ class SystemGPGEnvironment:
 
     def verify_detached(self,
                         signature_file: Path,
-                        data_file: Path,
+                        data_file: typing.IO[bytes],
                         require_all_good: bool = True,
                         ) -> OpenPGPSignatureList:
         """
         Verify the file against a detached signature
 
         Verify the data from data_file against the detached signature
-        from signature_file. Both files are specified by Path.
+        from signature_file.  data_file should be an open file,
+        whereas signature_file should be a Path object.
         Raise an exception if the verification fails.
 
         If require_all_good is True and the file contains multiple OpenPGP
@@ -305,7 +388,8 @@ class SystemGPGEnvironment:
 
         _, out, err = self._spawn_gpg(
             [GNUPG, "--batch", "--status-fd", "1", "--verify",
-             str(signature_file), str(data_file)])
+             str(signature_file), "-"],
+            stdin_file=data_file)
         return self._process_gpg_verify_output(out, err, require_all_good)
 
     def clear_sign_file(self, f, outf, keyid=None):
@@ -330,14 +414,18 @@ class SystemGPGEnvironment:
         outf.write(out.decode('utf8'))
 
     def _spawn_gpg(self, argv, stdin='', env_override={},
-                   raise_on_error=None):
+                   raise_on_error=None,
+                   stdin_file: typing.Optional[typing.IO[bytes]] = None):
         env = os.environ.copy()
         env['TZ'] = 'UTC'
         env.update(env_override)
 
+        if stdin_file is None:
+            stdin_file = subprocess.PIPE
+
         try:
             p = subprocess.Popen(argv,
-                                 stdin=subprocess.PIPE,
+                                 stdin=stdin_file,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  env=env)
@@ -373,29 +461,41 @@ class IsolatedGPGEnvironment(SystemGPGEnvironment):
     or use as a context manager (via 'with').
     """
 
-    __slots__ = ['_home', 'proxy']
-
-    def __init__(self, debug=False, proxy=None):
+    def __init__(self, debug=False, proxy=None, timeout=None):
         super().__init__(debug=debug)
         self.proxy = proxy
+        self.timeout = timeout
         self._home = tempfile.mkdtemp(prefix='gemato.')
 
         with open(os.path.join(self._home, 'dirmngr.conf'), 'w') as f:
             f.write(f'''# autogenerated by gemato
-
 # honor user's http_proxy setting
 honor-http-proxy
 
+# Disable automagically using Tor when running.  This often leads
+# to connectivity issues and is unexpected for users.
+no-use-tor
+
+# The default of "recursive-resolver" (see "man dirmngr") may cause
+# problems with corporate networks, where this is often prohibited.
+# It's better to setup the DNS resolver of your choice correctly
+# preferably with DNSSEC checks enabled and use the following setting:
+standard-resolver
+
 # enable debugging, in case we needed it
 log-file {os.path.join(self._home, 'dirmngr.log')}
 debug-level guru
 ''')
-        with open(os.path.join(self._home, 'gpg.conf'), 'w') as f:
-            f.write('''# autogenerated by gemato
 
-# we set validity directly on keys
-trust-model direct
-''')
+            if timeout is not None:
+                # GPG doesn't accept sub-second timeouts
+                gpg_timeout = math.ceil(timeout)
+                f.write(f"""
+# respect user-specified timeouts
+resolver-timeout {gpg_timeout}
+connect-timeout {gpg_timeout}
+""")
+
         with open(os.path.join(self._home, 'gpg-agent.conf'), 'w') as f:
             f.write(f'''# autogenerated by gemato
 
@@ -412,14 +512,15 @@ debug-level guru
             self.close()
 
     def clone(self):
-        return IsolatedGPGEnvironment(debug=self.debug, proxy=self.proxy)
+        return IsolatedGPGEnvironment(debug=self.debug, proxy=self.proxy,
+                                      timeout=self.timeout)
 
     def close(self):
         if self._home is not None:
             ret, sout, serr = self._spawn_gpg(
                 [GNUPGCONF, '--kill', 'all'])
             if ret != 0:
-                logging.warning(
+                LOGGER.warning(
                     f'{GNUPGCONF} --kill failed:\n'
                     f'{serr.decode("utf8", errors="backslashescape")}')
             if not self.debug:
@@ -428,8 +529,8 @@ debug-level guru
                     shutil.rmtree(self._home,
                                   onerror=_rmtree_error_handler)
             else:
-                logging.debug(f'GNUPGHOME left for debug purposes: '
-                              f'{self._home}')
+                LOGGER.debug(f'GNUPGHOME left for debug purposes: '
+                             f'{self._home}')
             self._home = None
 
     def import_key(self, keyfile, trust=True):
@@ -438,96 +539,60 @@ debug-level guru
             keyfile.read(),
             raise_on_error=OpenPGPKeyImportError)
 
-        if trust:
-            fprs = set()
-            for line in out.splitlines():
-                if line.startswith(b'[GNUPG:] IMPORT_OK'):
-                    fprs.add(line.split(b' ')[3].decode('ASCII'))
+        fprs = set()
+        for line in out.splitlines():
+            if line.startswith(b'[GNUPG:] IMPORT_OK'):
+                fprs.add(line.split(b' ')[3].decode('ASCII'))
+        if not fprs:
+            raise OpenPGPKeyImportError("No keys imported")
 
+        imported = self.list_keys(list(fprs))
+        missing = fprs - {fpr for fpr, uids in imported.items() if uids}
+        if missing:
+            raise OpenPGPKeyImportError(
+                "Import succeeded but no valid key for fingerprints: "
+                f"{missing}"
+            )
+
+        if trust:
+            self._trusted_keys.update(fprs)
             ownertrust = ''.join(f'{fpr}:6:\n' for fpr in fprs).encode('utf8')
             exitst, out, err = self._spawn_gpg(
                 [GNUPG, '--batch', '--import-ownertrust'],
                 ownertrust,
                 raise_on_error=OpenPGPKeyImportError)
 
-    def list_keys(self):
-        """
-        List fingerprints and UIDs of all keys in keyring
-
-        Returns a mapping from fingerprint (as a string) to an iterable
-        of UIDs.
-        """
-
-        exitst, out, err = self._spawn_gpg(
-            [GNUPG, '--batch', '--with-colons', '--list-keys'],
-            raise_on_error=OpenPGPKeyListingError)
-
-        prev_pub = None
-        fpr = None
-        ret = {}
-
-        for line in out.splitlines():
-            # were we expecting a fingerprint?
-            if prev_pub is not None:
-                if line.startswith(b'fpr:'):
-                    fpr = line.split(b':')[9].decode('ASCII')
-                    if not fpr.endswith(prev_pub):
-                        raise OpenPGPKeyListingError(
-                            f'Incorrect fingerprint {fpr} for key '
-                            f'{prev_pub}')
-                    logging.debug(
-                        f'list_keys(): fingerprint: {fpr}')
-                    ret[fpr] = []
-                    prev_pub = None
-                else:
-                    raise OpenPGPKeyListingError(
-                        f'No fingerprint in GPG output, instead got: '
-                        f'{line}')
-            elif line.startswith(b'pub:'):
-                # wait for the fingerprint
-                prev_pub = line.split(b':')[4].decode('ASCII')
-                logging.debug(f'list_keys(): keyid: {prev_pub}')
-            elif line.startswith(b'uid:'):
-                if fpr is None:
-                    raise OpenPGPKeyListingError(
-                        f'UID without key in GPG output: {line}')
-                uid = line.split(b':')[9]
-                _, addr = email.utils.parseaddr(
-                    uid.decode('utf8', errors='replace'))
-                if '@' in addr:
-                    logging.debug(f'list_keys(): UID: {addr}')
-                    ret[fpr].append(addr)
-                else:
-                    logging.debug(
-                        f'list_keys(): ignoring UID without mail: '
-                        f'{uid!r}')
-
-        return ret
-
     def refresh_keys_wkd(self):
         """
         Attempt to fetch updated keys using WKD.  Returns true if *all*
         keys were successfully found.  Otherwise, returns false.
         """
         if requests is None:
-            logging.debug('refresh_keys_wkd(): failing because requests'
-                          'module is missing')
+            LOGGER.debug('refresh_keys_wkd(): failing because requests'
+                         'module is missing')
             return False
 
         # list all keys in the keyring
         keys = self.list_keys()
         if not keys:
-            logging.debug('refresh_keys_wkd(): no keys found')
+            LOGGER.debug('refresh_keys_wkd(): no keys found')
             return False
         addrs = set()
         for key, uids in keys.items():
-            if not uids:
-                logging.debug(
-                    f'refresh_keys_wkd(): failing due to no UIDs on '
-                    f'key {key}')
+            key_addrs = {
+                addr for uid in uids
+                if "@" in (addr := email.utils.parseaddr(
+                    uid.decode("utf8", errors="replace")
+                )[1])
+            }
+            print(key_addrs)
+            if not key_addrs:
+                LOGGER.debug(
+                    f"refresh_keys_wkd(): failing due to no email addresses "
+                    f"on key {key}")
                 return False
-            addrs.update(uids)
-        keys = set(keys)
+            addrs.update(key_addrs)
+        expected_keys = frozenset(keys)
 
         data = b''
         proxies = {}
@@ -539,13 +604,13 @@ debug-level guru
         for a in addrs:
             url = get_wkd_url(a)
             try:
-                resp = requests.get(url, proxies=proxies)
+                resp = requests.get(url, proxies=proxies, timeout=self.timeout)
                 resp.raise_for_status()
             except (requests.exceptions.ConnectionError,
                     requests.exceptions.HTTPError,
                     ) as e:
-                logging.debug(f'refresh_keys_wkd(): failing due to failed '
-                              f'request for {url}: {e}')
+                LOGGER.debug(f'refresh_keys_wkd(): failing due to failed '
+                             f'request for {url}: {e}')
                 return False
             data += resp.content
 
@@ -554,23 +619,37 @@ debug-level guru
             data,
             raise_on_error=OpenPGPKeyRefreshError)
 
-        # we need to explicitly ensure all keys were fetched
+        imported_keys = set()
         for line in out.splitlines():
             if line.startswith(b'[GNUPG:] IMPORT_OK'):
                 fpr = line.split(b' ')[3].decode('ASCII')
-                logging.debug(
+                LOGGER.debug(
                     f'refresh_keys_wkd(): import successful for key: {fpr}')
-                if fpr in keys:
-                    keys.remove(fpr)
-                else:
-                    # we need to delete unexpected keys
-                    exitst, out, err = self._spawn_gpg(
-                        [GNUPG, '--batch', '--delete-keys', fpr],
-                        raise_on_error=OpenPGPKeyRefreshError)
-        if keys:
-            logging.debug(
+                imported_keys.add(fpr)
+
+        # Need to explicitly ensure all keys were fetched
+        # However:
+        # - any key MAY appear 0 or more times.
+        # - expected keys SHOULD be present.
+        # - unexpected keys MAY also be present.
+        unexpected_keys = imported_keys.difference(expected_keys)
+        if unexpected_keys:
+            # we need to delete unexpected keys
+            LOGGER.debug(
+                f'refresh_keys_wkd(): got unexpected key, will remove: '
+                f'{unexpected_keys}')
+            # 128x 40-byte fingerprints = 5KiB commandline max
+            # If this contains a lot of keys, it should just blow up, but that
+            # saves complexity.
+            exitst, out, err = self._spawn_gpg(
+                [GNUPG, '--batch', '--delete-keys'] + list(unexpected_keys),
+                raise_on_error=OpenPGPKeyRefreshError)
+
+        not_updated_keys = expected_keys.difference(imported_keys)
+        if not_updated_keys:
+            LOGGER.debug(
                 f'refresh_keys_wkd(): failing due to non-updated keys: '
-                f'{keys}')
+                f'{not_updated_keys}')
             return False
 
         return True
@@ -580,13 +659,31 @@ debug-level guru
         if keyserver is not None:
             ks_args = ['--keyserver', keyserver]
 
+        expected_fprs = set(self.list_keys())
         exitst, out, err = self._spawn_gpg(
-            [GNUPG, '--batch', '--refresh-keys'] + ks_args,
+            [GNUPG, '--batch', '--refresh-keys', '--status-fd', '1'] + ks_args,
             raise_on_error=OpenPGPKeyRefreshError)
 
+        imported_fprs = set()
+        for line in out.splitlines():
+            if line.startswith(b'[GNUPG:] IMPORT_OK'):
+                imported_fprs.add(line.split(b' ')[3].decode('ASCII'))
+
+        if imported_fprs != expected_fprs:
+            extra_keys = imported_fprs - expected_fprs
+            missing_keys = expected_fprs - imported_fprs
+            if extra_keys:
+                raise OpenPGPKeyRefreshError(
+                    f"Keyserver update injected additional keys: {extra_keys}"
+                )
+            if missing_keys:
+                raise OpenPGPKeyRefreshError(
+                    f"Keyserver is missing keys: {missing_keys}"
+                )
+
     def refresh_keys(self, allow_wkd=True, keyserver=None):
-        logging.debug(f'refresh_keys(allow_wkd={allow_wkd}, '
-                      f'keyserver={keyserver}) called')
+        LOGGER.debug(f'refresh_keys(allow_wkd={allow_wkd}, '
+                     f'keyserver={keyserver}) called')
 
         if allow_wkd and self.refresh_keys_wkd():
             return
@@ -612,7 +709,7 @@ class PGPyEnvironment:
 
     __slots__ = ['debug', 'keyring', 'proxy']
 
-    def __init__(self, debug=False, proxy=None):
+    def __init__(self, debug=False, proxy=None, timeout=None):
         if pgpy is None:
             raise OpenPGPNoImplementation('install PGPy')
         self.debug = debug
@@ -649,12 +746,12 @@ class PGPyEnvironment:
                     try:
                         verifies = k.parent.verify(k)
                     except pgpy.errors.PGPError:
-                        logging.debug(
+                        LOGGER.debug(
                             f'Rejecting subkey {fpr} due to missing sig')
                         self.keyring.unload(k)
                     else:
                         if not verifies:
-                            logging.debug(
+                            LOGGER.debug(
                                 f'Rejecting subkey {fpr} since parent '
                                 f'key signature does not check out')
                             self.keyring.unload(k)
